import re
import zlib

import numpy as np
from antlr4 import *
from antlr4.error.ErrorListener import ErrorListener
from dateutil import parser as dateparser
from textwrap import dedent

from norm import config
from norm.executable import NormExecutable
from norm.executable.command import Command
from norm.executable.schema.declaration import *
from norm.executable.expression.arithmetic import *
from norm.executable.expression.evaluation import *
from norm.executable.expression.query import *
from norm.executable.expression.slice import *
from norm.executable.expression.code import *
from norm.executable.schema.implementation import *
from norm.executable.schema.type import *
from norm.executable.schema.namespace import *
from norm.executable.schema.variable import UnquoteVariable
from norm.grammar.literals import AOP, COP, LOP, ImplType, ConstantType, MOP
from norm.grammar.normLexer import normLexer
from norm.grammar.normListener import normListener
from norm.grammar.normParser import normParser


class QType(Enum):
    FOREACH = 0
    FORANY = 1
    EXIST = 2
    NEXIST = 3


class QuantifiedLambda(object):

    def __init__(self, expr: NormExpression):
        self.expr = expr
        self.lam = expr.lam
        self.cols = []
        self.quantifiers = []

    def __contains__(self, item):
        return item in self.lam

    @property
    def VAR_OID(self):
        return self.lam.VAR_OID

    @property
    def cloned_from(self):
        return self.lam.cloned_from

    @property
    def name(self):
        return self.lam.name

    @property
    def is_functional(self):
        return self.lam.is_functional

    @property
    def atomic(self):
        return self.lam.atomic

    @property
    def nargs(self):
        return self.lam.nargs

    def fill_time(self, df):
        if self.lam.VAR_TIMESTAMP not in df.columns:
            df[self.lam.VAR_TIMESTAMP] = np.datetime64(datetime.utcnow())
        return df

    def fill_oid(self, df):
        var_oid = self.lam.VAR_OID
        if df.index.name == var_oid:
            return df

        if var_oid in df.columns:
            return df.set_index(var_oid)

        c = None
        for col in self.cols:
            if c is None:
                c = df[col].astype(str)
            else:
                c = c.str.cat(df[col].astype(str))
        df[var_oid] = c.str.encode('utf-8').apply(zlib.adler32).astype('int64')
        df = df.set_index(var_oid)
        return df

    @property
    def variables(self):
        return self.lam.variables

    @property
    def data(self):
        if len(self.cols) > 0:
            return self.lam.data.groupby(self.cols)
        else:
            return self.lam.data

    def get_type(self, name):
        return self.lam.get_type(name)

    def get_grouped_variables(self):
        from norm.models import lambdas
        variables = dict((v.name, v) for v in self.lam.variables)
        return [variables.get(col, lambdas.Any) for col in self.cols]

    def add_cols(self, qtype, cols):
        pos = len(self.cols)
        self.cols.extend(cols)
        self.quantifiers.append((qtype, pos, pos + len(cols)))

    def execute(self, context):
        return self.expr.execute(context)

    def quantify(self, df):
        if len(self.cols) == 0 or len(df) == 0:
            return df
        # assert(all([col in df.columns for col in self.cols]))
        while len(self.quantifiers) > 0:
            qtype, begin, end = self.quantifiers.pop()
            if len(df) == 0:
                continue
            if qtype == QType.EXIST:
                if begin == 0:
                    df = df.iloc[:1]
                else:
                    df = df.groupby(self.cols[:begin]).min().reset_index()
                    df.index.name = self.lam.VAR_OID
                    df[self.lam.VAR_TIMESTAMP] = np.datetime64(datetime.utcnow())
            elif qtype == QType.NEXIST:
                if begin == 0:
                    if not df.empty:
                        df = df.iloc[:1]
                        df[self.lam.VAR_LABEL] = 0.0
                else:
                    df = df[df[self.cols[begin]].isna()]
                    df.index.name = self.lam.VAR_OID
                    df[self.lam.VAR_TIMESTAMP] = np.datetime64(datetime.utcnow())
            elif qtype == QType.FORANY:
                total = len(self.lam.data[self.cols[begin:end]].drop_duplicates())
                if begin == 0:
                    df = df if len(df) == total else df.iloc[0:0]
                else:
                    df['_tmp_count'] = df.groupby(self.cols[:begin])[self.cols[begin]].transform('count')
                    df = df[df['_tmp_count'] == total].drop(columns=['_tmp_count'])
        return df


class ParseError(ValueError):
    pass


class NormErrorListener(ErrorListener):

    def __init__(self):
        super(NormErrorListener, self).__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        err_msg = "line " + str(line) + ":" + str(column) + " " + msg
        raise ParseError(err_msg)


error_listener = NormErrorListener()
walker = ParseTreeWalker()


class NormCompiler(normListener):
    TMP_VARIABLE_STUB = 'tmp_'
    THAT_VARIABLE_NAME = 'that'

    def __init__(self, context_id, user, session=None):
        self.context_id = context_id
        self.scopes = []
        self.stack = []
        self.session = session
        self.user = user
        self.context_namespace = None
        self.user_namespace = None
        self.search_namespaces = None
        self.set_namespaces()
        self._python_context = {}
        self.that = None

    def set_namespaces(self):
        if self.context_id:
            self.context_namespace = '{}.{}'.format(config.CONTEXT_NAMESPACE_STUB, self.context_id)
        if self.user:
            self.user_namespace = '{}.{}'.format(config.USER_NAMESPACE_STUB, self.user.username)
        from norm.models import NativeLambda, CoreLambda
        self.search_namespaces = [NativeLambda.NAMESPACE, CoreLambda.NAMESPACE, self.context_namespace,
                                  self.user_namespace]

    def get_scope(self, name):
        for scope, scope_lex in reversed(self.scopes):
            if name in scope:
                return scope
        return None

    @property
    def scope(self):
        return self.scopes[-1][0] if len(self.scopes) > 0 else None

    @property
    def scope_lex(self):
        return self.scopes[-1][1] if len(self.scopes) > 0 else None

    def _push(self, exe):
        self.stack.append(exe)

    def _pop(self):
        """
        :rtype: NormExecutable
        """
        return self.stack.pop()

    def _peek(self):
        """
        :rtype: NormExecutable
        """
        return self.stack[-1] if len(self.stack) > 0 else None

    def optimize(self, exe):
        """
        Optimize the AST to have a more efficient execution plan
        # TODO: optimization strategies
        * Filtering conditions can be combined and executed in batch instead of sequential
        * Arithmetic equations can be combined and passed to DF in batch instead of sequential
        :param exe: the executable to be optimized
        :type exe: NormExecutable
        :return: the optimized executable
        :rtype: NormExecutable
        """
        return exe

    def parse(self, script):
        if script is None or not isinstance(script, str):
            return None
        script = script.strip(' \r\n\t')
        if script == '':
            return None
        try:
            lexer = normLexer(InputStream(script))
            stream = CommonTokenStream(lexer)
            parser = normParser(stream)
            parser.addErrorListener(error_listener)
            return parser.script()
        except ParseError as e:
            msg = 'Can not parse {}'.format(script)
            logger.error(msg)
            logger.error(e)
            raise e

    def compile(self, script):
        tree = self.parse(script)
        if tree is not None:
            walker.walk(self, tree)
            return True
        else:
            return False
        # return self.optimize(self._pop())

    def temp_lambda(self, variables):
        return Lambda(self.context_namespace, self.TMP_VARIABLE_STUB + str(uuid.uuid4()), variables=variables)

    @property
    def python_context(self):
        return self._python_context

    @python_context.setter
    def python_context(self, python_context=None):
        if python_context is None:
            self._python_context = {}
        else:
            self._python_context.update(python_context)

    def execute(self, script):
        self.stack = []
        self.scopes = []
        if not self.compile(dedent(script)):
            return None

        results = None
        exe = None
        while len(self.stack) > 0:
            exe = self._pop()
            if isinstance(exe, NormExecutable):
                results = exe.execute(self)
            else:
                # TODO: shouldn't be here?
                results = exe
            if isinstance(results, Index) and isinstance(exe, NormExpression) and exe.lam is not None:
                results = exe.lam.data.loc[results]
        if isinstance(results, DataFrame):
            results.columns = results.columns.str.replace(NormExecutable.VARIABLE_SEPARATOR, '.')
        elif isinstance(results, Series):
            results = DataFrame(data={results.name.replace(NormExecutable.VARIABLE_SEPARATOR, '.'): results})
        if exe is not None and isinstance(exe, NormExecutable):
            self.that = exe.lam
        return results

    def exitStatement(self, ctx: normParser.StatementContext):
        if ctx.typeDeclaration():
            if ctx.IMPL():
                query = self._pop()  # type: NormExpression
                type_declaration = self._pop()  # type: TypeDeclaration
                description = self._pop() if ctx.comments() else ''
                type_declaration.description = description
                op = ImplType(ctx.IMPL().getText())
                self._push(TypeImplementation(type_declaration, op, query, description).compile(self))
            elif ctx.comments():
                # otherwise just pass through
                type_declaration = self._pop()  # type: TypeDeclaration
                description = self._pop()
                type_declaration.description = description  # need to be recompiled
                self._push(type_declaration.compile(self))
        elif ctx.argumentDeclarations():
            args = self._pop()  # type: List[ArgumentDeclaration]
            type_ = self._pop()  # type: TypeName
            self._push(AdditionalTypeDeclaration(type_, args).compile(self))
        elif ctx.renames():
            renames = self._pop()  # type: List[RenameArgument]
            type_ = self._pop()  # type: TypeName
            self._push(RenameTypeDeclaration(type_, renames).compile(self))
        elif ctx.codeExpression():
            code: CodeExpr = self._pop()
            type_: TypeName = self._pop()
            description = self._pop() if ctx.comments() else ''
            op = ImplType(ctx.IMPL().getText())
            if type_.lam is None and op == ImplType.DEF:
                self._push(CodeTypeDeclaration(type_, code.code, description).compile(self))
            else:
                self._push(TypeImplementation(type_, op, code, description).compile(self))
        elif ctx.imports() or ctx.exports() or ctx.commands() or ctx.multiLineExpression():
            if ctx.comments():
                expr = self._pop()
                # ignore comments
                self._pop()
                self._push(expr)

    def exitNone(self, ctx: normParser.NoneContext):
        self._push(Constant(ConstantType.NULL, None))

    def exitBool_c(self, ctx: normParser.Bool_cContext):
        self._push(Constant(ConstantType.BOOL, ctx.getText().lower() == 'true'))

    def exitInteger_c(self, ctx: normParser.Integer_cContext):
        self._push(Constant(ConstantType.INT, int(ctx.getText())))

    def exitFloat_c(self, ctx: normParser.Float_cContext):
        self._push(Constant(ConstantType.FLT, float(ctx.getText())))

    def exitString_c(self, ctx: normParser.String_cContext):
        self._push(Constant(ConstantType.STR, str(ctx.getText()[1:-1])))

    def exitPattern(self, ctx: normParser.PatternContext):
        try:
            self._push(Constant(ConstantType.PTN, re.compile(str(ctx.getText()[2:-1]))))
        except:
            raise ParseError('Pattern constant {} is in wrong format, should be Python regex pattern'
                             .format(ctx.getText()))

    def exitUuid(self, ctx: normParser.UuidContext):
        self._push(Constant(ConstantType.UID, str(ctx.getText()[2:-1])))

    def exitUrl(self, ctx: normParser.UrlContext):
        self._push(Constant(ConstantType.URL, str(ctx.getText()[2:-1])))

    def exitDatetime(self, ctx: normParser.DatetimeContext):
        self._push(Constant(ConstantType.DTM, dateparser.parse(ctx.getText()[2:-1], fuzzy=True)))

    def exitConstant(self, ctx: normParser.ConstantContext):
        if ctx.LSBR() or ctx.LBR():
            constants = list(reversed([self._pop() for ch in ctx.children
                                       if isinstance(ch, normParser.ConstantContext)]))  # type: List[Constant]
            if ctx.LBR():
                self._push(TupleConstant(tuple(constant.type_ for constant in constants),
                                         tuple(constant.value for constant in constants)))
            else:
                types = set(constant.type_ for constant in constants)
                if len(types) > 1:
                    type_ = ConstantType.ANY
                else:
                    type_ = types.pop()
                self._push(ListConstant(type_, [constant.value for constant in constants]))

    def exitQueryProjection(self, ctx: normParser.QueryProjectionContext):
        variables = list(reversed([self._pop() for ch in ctx.children
                                   if isinstance(ch, normParser.VariableContext)]))
        self._push(Projection(variables))

    def exitComments(self, ctx: normParser.CommentsContext):
        spaces = ' \r\n\t'
        cmt = ctx.getText()
        if ctx.MULTILINE():
            cmt = cmt.strip(spaces)[2:-2].strip(spaces)
        elif ctx.SINGLELINE():
            cmt = '\n'.join(cmt_line.strip(spaces)[2:].strip(spaces) for cmt_line in cmt.split('\n'))
        self._push(cmt)

    def exitImports(self, ctx: normParser.ImportsContext):
        type_: TypeName = self._pop() if ctx.typeName() else None
        namespace = [str(v) for v in ctx.VARNAME()]
        variable = namespace.pop() if ctx.AS() else None
        self._push(Import('.'.join(namespace), type_, variable).compile(self))

    def exitExports(self, ctx: normParser.ExportsContext):
        type_: TypeName = self._pop()
        namespace = [str(v) for v in ctx.VARNAME()]
        variable = namespace.pop() if ctx.AS() else None
        self._push(Export('.'.join(namespace), type_, variable).compile(self))

    def exitCommands(self, ctx: normParser.CommandsContext):
        type_: TypeName = self._pop()
        op = MOP(ctx.SPACED_COMMAND().getText().strip().lower())
        self._push(Command(op, type_).compile(self))

    def exitArgumentDeclaration(self, ctx: normParser.ArgumentDeclarationContext):
        variable_property = ctx.argumentProperty().getText() if ctx.argumentProperty() else None
        optional = variable_property is not None and variable_property.lower().find('optional') >= 0
        as_oid = variable_property is not None and variable_property.lower().find('oid') >= 0
        as_time = variable_property is not None and variable_property.lower().find('time') >=0
        type_name: TypeName = self._pop()
        variable_name: VariableName = self._pop()
        self._push(ArgumentDeclaration(variable_name, type_name, optional, as_oid, as_time))

    def enterArgumentDeclarations(self, ctx: normParser.ArgumentDeclarationsContext):
        type_name = self._peek()
        if type_name is not None and isinstance(type_name, TypeName) and type_name.lam is not None:
            self.scopes.append((type_name.lam, 'argument_declarations'))

    def exitArgumentDeclarations(self, ctx: normParser.ArgumentDeclarationsContext):
        args = list(reversed([self._pop() for ch in ctx.children
                              if isinstance(ch, normParser.ArgumentDeclarationContext)]))
        self._push(args)
        if self.scope_lex == 'argument_declarations':
            self.scopes.pop()

    def exitRename(self, ctx: normParser.RenameContext):
        new_name: VariableName = self._pop()
        original_name: VariableName = self._pop()
        self._push(RenameArgument(original_name.name, new_name.name))

    def exitRenames(self, ctx: normParser.RenamesContext):
        args = list(reversed([self._pop() for ch in ctx.children
                              if isinstance(ch, normParser.RenameContext)]))
        self._push(args)
        if self.scope_lex == 'renames':
            self.scopes.pop()

    def enterRenames(self, ctx: normParser.RenamesContext):
        type_name = self._peek()
        if type_name is not None and isinstance(type_name, TypeName) and type_name.lam is not None:
            self.scopes.append((type_name.lam, 'renames'))

    def exitTypeDeclaration(self, ctx: normParser.TypeDeclarationContext):
        output_type_name = self._pop() if ctx.typeName(1) else None  # type: TypeName
        args = self._pop() if ctx.argumentDeclarations() else None  # type: List[ArgumentDeclaration]
        type_name = self._pop()  # type: TypeName
        self._push(TypeDeclaration(type_name, args, output_type_name).compile(self))

    def exitTypeName(self, ctx: normParser.TypeNameContext):
        typename = ctx.VARNAME()
        if typename:
            version = ctx.version().getText() if ctx.version() else None
            self._push(TypeName(str(typename), version).compile(self))
        elif ctx.LSBR():
            intern = self._pop()  # type: TypeName
            self._push(ListType(intern).compile(self))
        else:
            raise ParseError('Not a valid type name definition')

    def exitUnquote_variable(self, ctx:normParser.Unquote_variableContext):
        unqoted_variables = []
        template = []
        for ch in ctx.children:
            if isinstance(ch, normParser.VariableContext):
                variable: VariableName = self._pop()
                unqoted_variables.append(variable)
            else:
                template.append(ch.getText())
        self._push(UnquoteVariable(''.join(template), list(reversed(unqoted_variables))).compile(self))

    def exitVariable(self, ctx: normParser.VariableContext):
        if ctx.VARNAME():
            name = ctx.VARNAME().getText()
        elif ctx.COMMAND():
            name = ctx.COMMAND().getText()
        elif ctx.ARGOPT():
            name = ctx.ARGOPT().getText()
        else:
            name = None

        if ctx.DOT():
            # chained variable
            v: VariableName = self._pop()
            scope: VariableName = self._pop()
            self._push(VariableName(scope, v.name).compile(self))
        elif name is not None:
            self._push(VariableName(None, name).compile(self))

    def exitArgumentExpression(self, ctx: normParser.ArgumentExpressionContext):
        if isinstance(self._peek(), ArgumentExpr):
            return

        projection = self._pop() if ctx.queryProjection() else None  # type: Projection
        expr = self._pop() if ctx.arithmeticExpression() else None  # type: ArithmeticExpr
        op = COP(ctx.spacedConditionOperator().conditionOperator().getText().lower()) \
            if ctx.spacedConditionOperator() else None  # type: COP
        # = is treated as op is None
        variable = self._pop() if ctx.variable() else None  # type: VariableName
        if variable is None and projection is None and op is None and expr is not None \
                and isinstance(expr, EvaluationExpr) and isinstance(expr.variable, ColumnVariable) \
                and len(expr.args) == 0 and expr.projection is not None:
            self._push(ArgumentExpr(expr.variable, op, None, expr.projection).compile(self))
        else:
            self._push(ArgumentExpr(variable, op, expr, projection).compile(self))

    def enterArgumentExpressions(self, ctx: normParser.ArgumentExpressionsContext):
        expr = self._peek()
        if expr is not None and isinstance(expr, (VariableName, EvaluationExpr)) and expr.lam is not None:
            self.scopes.append((expr.lam, 'argument_expressions'))

    def exitArgumentExpressions(self, ctx: normParser.ArgumentExpressionsContext):
        args = list(reversed([self._pop() for ch in ctx.children
                              if isinstance(ch, normParser.ArgumentExpressionContext)]))
        self._push(args)
        if self.scope_lex == 'argument_expressions':
            self.scopes.pop()

    def enterMultiLineExpression(self, ctx: normParser.MultiLineExpressionContext):
        type_declaration = self._peek()
        if type_declaration is not None and isinstance(type_declaration, TypeDeclaration) \
                and type_declaration.lam is not None:
            self.scopes.append((type_declaration.lam, 'multiline'))

    def exitContext(self, ctx:normParser.ContextContext):
        if ctx.WITH():
            expr: NormExpression = self._pop()
            self.scopes.append((QuantifiedLambda(expr), 'one_line_context'))
        elif ctx.FOREACH() or ctx.FORANY() or ctx.EXIST() or ctx.NEXIST():
            if self.scope is not None and isinstance(self.scope, QuantifiedLambda):
                lam = self.scope
            else:
                lam = None
            tns = []
            for ch in ctx.children:
                if isinstance(ch, normParser.VariableContext):
                    tn = self._pop()
                    #assert(isinstance(tn, ColumnVariable))
                    tns.append(tn.name)
                    if lam is None:
                        lam = tn.lam
                    #else:
                    #    assert(tn.lam is lam)
            if ctx.FOREACH():
                qtype = QType.FOREACH
            elif ctx.FORANY():
                qtype = QType.FORANY
            elif ctx.EXIST():
                qtype = QType.EXIST
            else:
                qtype = QType.NEXIST
            lam.add_cols(qtype, list(reversed(tns)))
            if lam is not self.scope:
                self.scopes.append((lam, 'one_line_context'))

    def exitContextualOneLineExpression(self, ctx:normParser.ContextualOneLineExpressionContext):
        while self.scope_lex == 'one_line_context':
            scope, lex = self.scopes.pop()
            if isinstance(scope, QuantifiedLambda):
                expr = self._pop()  # type: NormExpression
                self._push(QuantifiedQueryExpr(expr, scope).compile(self))

    def exitMultiLineExpression(self, ctx: normParser.MultiLineExpressionContext):
        if ctx.newlineLogicalOperator():
            expr2 = self._pop()  # type: NormExpression
            expr1 = self._pop()  # type: NormExpression
            op = LOP.parse(ctx.newlineLogicalOperator().logicalOperator().getText())
            self._push(QueryExpr(op, expr1, expr2).compile(self))
        if self.scope_lex == 'multiline':
            self.scopes.pop()

    def enterSpacedLogicalOperator(self, ctx:normParser.SpacedLogicalOperatorContext):
        expr = self._peek()
        self.scopes.append((expr.lam, 'lop'))

    def enterNewlineLogicalOperator(self, ctx:normParser.NewlineLogicalOperatorContext):
        expr = self._peek()
        self.scopes.append((expr.lam, 'lop'))

    def exitOneLineExpression(self, ctx: normParser.OneLineExpressionContext):
        if ctx.queryProjection():
            projection: Projection = self._pop()
            expr = self._pop()
            if isinstance(expr, VariableName):
                expr = EvaluationExpr([], expr)
            elif isinstance(expr, Constant):
                expr = ConstantAssignmentExpr(expr)

            if projection.with_unquote:
                expr = PivotQueryExpr(expr, projection)
            else:
                expr.projection = projection
            self._push(expr.compile(self))
        elif ctx.NOT():
            expr = self._pop()  # type: NormExpression
            self._push(NegatedQueryExpr(expr).compile(self))
        elif ctx.spacedLogicalOperator():
            expr2 = self._pop()  # type: NormExpression
            expr1 = self._pop()  # type: NormExpression
            op = LOP.parse(ctx.spacedLogicalOperator().logicalOperator().getText())
            self._push(QueryExpr(op, expr1, expr2).compile(self))
            if self.scope_lex == 'lop':
                self.scopes.pop()
        elif isinstance(self._peek(), TupleConstant):
            expr = self._pop()
            lam = self.scope
            values = expr.value
            from norm.models.norm import Lambda
            from norm.models import lambdas
            variables = [v.name for v in lam.variables]
            nanonymous = len(values) - len(variables)
            variables.extend('{}{}'.format(Lambda.VAR_ANONYMOUS_STUB, i) for i in range(nanonymous))
            data = dict((name, value) for name, value in zip(variables, values))
            self._push(AddDataEvaluationExpr(lam, data, False))

    def exitConditionExpression(self, ctx: normParser.ConditionExpressionContext):
        if ctx.spacedConditionOperator():
            qexpr = self._pop()  # type: ArithmeticExpr
            aexpr = self._pop()  # type: ArithmeticExpr
            cop = COP(ctx.spacedConditionOperator().conditionOperator().getText().lower())
            self._push(ConditionExpr(cop, aexpr, qexpr).compile(self))

    def exitArithmeticExpression(self, ctx: normParser.ArithmeticExpressionContext):
        if ctx.slicedExpression():
            return

        op = None
        if ctx.MOD():
            op = AOP.MOD
        elif ctx.EXP():
            op = AOP.EXP
        elif ctx.TIMES():
            op = AOP.MUL
        elif ctx.DIVIDE():
            op = AOP.DIV
        elif ctx.PLUS():
            op = AOP.ADD
        elif ctx.MINUS():
            op = AOP.SUB

        if op is not None:
            expr2 = self._pop()  # type: ArithmeticExpr
            if isinstance(expr2, ArgumentExpr):
                expr2 = expr2.variable
            expr1 = self._pop() if ctx.arithmeticExpression(1) else None  # type: ArithmeticExpr
            if isinstance(expr1, AddDataEvaluationExpr):
                expr1 = list(expr1.data.values())[0]
                assert (isinstance(expr1, ArithmeticExpr))
            self._push(ArithmeticExpr(op, expr1, expr2).compile(self))

    def exitSlicedExpression(self, ctx: normParser.SlicedExpressionContext):
        if ctx.LSBR():
            if ctx.evaluationExpression(1):
                expr_range = self._pop()  # type: NormExpression
                expr = self._pop()  # type: NormExpression
                self._push(EvaluatedSliceExpr(expr, expr_range).compile(self))
            else:
                end = self._pop() if ctx.integer_c(1) else None  # type: Constant
                start = self._pop() if ctx.integer_c(0) else None  # type: Constant
                expr = self._pop()
                start_value = start.value
                if end is not None:
                    end_value = end.value
                elif ctx.COLON():
                    colon_before_number = True
                    for ch in ctx.children:
                        if ch == ctx.COLON():
                            break
                        elif ch == ctx.integer_c(0):
                            colon_before_number = False
                            break
                    if colon_before_number:
                        end_value = start_value
                        start_value = 0
                    else:
                        end_value = None
                else:
                    end_value = start_value + 1
                self._push(SliceExpr(expr, start_value, end_value).compile(self))

    def exitEvaluationExpression(self, ctx: normParser.EvaluationExpressionContext):
        if ctx.DOT():
            rexpr = self._pop()
            lexpr = self._pop()
            self._push(ChainedEvaluationExpr(lexpr, rexpr).compile(self))
        elif ctx.argumentExpressions():
            args = self._pop() if ctx.argumentExpressions() else []  # type: List[ArgumentExpr]
            variable = self._pop() if ctx.variable() else None  # type: VariableName
            self._push(EvaluationExpr(args, variable, None).compile(self))

    def exitCodeExpression(self, ctx: normParser.CodeExpressionContext):
        self._push(CodeExpr(ctx.code().getText()).compile(self))



import uuid
from enum import Enum

from norm.utils import hash_df
from pandas import DataFrame, Series, Index

from norm.grammar.literals import COP
from norm.executable import NormError, Projection
from norm.executable.expression.argument import ArgumentExpr
from norm.executable.schema.variable import VariableName, ColumnVariable, JoinVariable, UnquoteVariable
from norm.executable.expression import NormExpression

from collections import OrderedDict

from typing import List, Dict

import logging
logger = logging.getLogger(__name__)


class EvaluationExpr(NormExpression):

    def __init__(self, args, variable=None, projection=None):
        """
        The evaluation of an expression either led by a variable name
        :param args: the arguments provided
        :type args: List[ArgumentExpr]
        :param variable: the variable name to evaluate
        :type variable: VariableName
        :type projection: Projection
        """
        super().__init__()
        self.variable: VariableName = variable
        self.args: List[ArgumentExpr] = args
        self.inputs: Dict[str, DataFrame] = None
        self.outputs: Dict[str, str] = None
        self.join_variables: List[JoinVariable] = []
        self.projection: Projection = projection
        self.equalities: Dict[str, str] = OrderedDict()
        from norm.models.norm import Lambda
        self.equality_scope: Lambda = None

    @property
    def is_constant_arguments(self):
        return all(arg.is_constant for arg in self.args)

    def check_assignment_arguments(self):
        """
        Check whether the arguments are in the style of assignments. Ex. 234, 'asfs', b=3, c=5
        :return: True or False
        :rtype: bool
        """
        keyword_arg = False  # positional argument
        for arg in self.args:
            if arg.op is None:
                if arg.variable is None:
                    if keyword_arg:
                        msg = 'Keyword based arguments should come after positional arguments'
                        logger.error(msg)
                        logger.error([str(arg) for arg in self.args])
                        raise NormError(msg)
                else:
                    keyword_arg = True
            else:
                return False
        return True

    def equality_arg(self, arg):
        """
        Check whether the argument is referring to an equality to the previous context or not
        :param arg: the argument expr to check
        :type arg: ArgumentExpr
        :return: True if it is
        :rtype: bool
        """
        if (arg.is_assign_operator or arg.op is COP.EQ) and isinstance(arg.expr, ColumnVariable) and \
                isinstance(arg.variable, ColumnVariable):
            self.equalities[arg.expr.name] = arg.variable.name
            if self.equality_scope is None:
                self.equality_scope = arg.expr.lam
            else:
                assert (self.equality_scope is arg.expr.lam)
            return True
        else:
            return False

    def build_assignment_inputs(self, lam):
        """
        If the arguments are in assignment style, build the inputs as assignments
        :param lam: given the Lambda to be evaluated
        :return: a dictionary mapping from a variable to an expression for the inputs
        :rtype: Dict
        """
        if lam.nargs == 0:
            # Lambda is anonymously created
            from norm.models.norm import Lambda
            return OrderedDict((arg.variable.name if arg.variable else '{}{}'.format(Lambda.VAR_ANONYMOUS_STUB, i),
                                arg.expr) for i, arg in enumerate(self.args))

        keyword_arg = False
        inputs = OrderedDict()
        from norm.models.norm import Variable
        for ov, arg in zip(lam.variables, self.args):  # type: Variable, ArgumentExpr
            if self.equality_arg(arg):
                continue
            if arg.expr is None:
                continue
            if arg.op is None and arg.variable is not None:
                keyword_arg = True
            if not keyword_arg:
                inputs[ov.name] = arg.expr
            else:
                inputs[arg.variable.name] = arg.expr
        return inputs

    def build_conditional_inputs(self):
        """
        If the arguments are in conditional style, build the inputs as conditionals
        :return: a query string
        :rtype: Dict
        """
        inputs = []
        for arg in self.args:
            if self.equality_arg(arg):
                continue
            if arg.op is None or arg.expr is None:
                continue
            vn = arg.variable.name
            if isinstance(arg.variable, JoinVariable):
                self.join_variables.append(arg.variable)
                vn = str(arg.variable)
            if arg.op == COP.LK:
                condition = '{}.str.contains({})'.format(vn, arg.expr)
            else:
                condition = '{} {} ({})'.format(vn, arg.op, arg.expr)
            inputs.append(condition)
        return ' and '.join(inputs)

    def build_outputs(self, lam):
        """
        Build the outputs according to the projection
        :return: a dictionary mapping from the original variable to the new variable
        :rtype: Dict
        """
        from norm.models.norm import Lambda
        outputs = OrderedDict()
        from norm.models.norm import Variable
        for ov, arg in zip(lam.variables, self.args):  # type: Variable, ArgumentExpr
            if arg.projection is not None:
                assert(len(arg.projection.variables) <= 1)
                if arg.variable is None:
                    outputs[ov.name] = arg.projection.variables[0].name
                else:
                    if len(arg.projection.variables) == 0:
                        outputs[arg.variable.name] = arg.variable.name
                    else:
                        outputs[arg.variable.name] = arg.projection.variables[0].name
        if self.projection is not None:
            assert(len(self.projection.variables) <= 1)
            if self.projection.num == 1:
                new_lambda_name = self.projection.variables[0].name
            else:
                new_lambda_name = lam.name
            outputs = OrderedDict((key, self.VARIABLE_SEPARATOR.join([new_lambda_name, value])) for key, value
                                  in outputs.items())
            if lam.is_functional:
                outputs[lam.VAR_OUTPUT] = new_lambda_name
            else:
                outputs[lam.VAR_OID] = new_lambda_name
        return outputs

    def compile(self, context):
        if self.variable is None:
            from norm.models.norm import Lambda
            assert(context.scope is not None)
            assert(isinstance(context.scope, Lambda))
            lam = context.scope
            # constructing new objects
            assert(self.check_assignment_arguments())
            self.inputs = self.build_assignment_inputs(lam)
            self.outputs = None
            is_to_add_data = True
        else:
            lam = self.variable.lam
            if lam is None:
                # Might need to be combined with previous context
                return self
            if self.check_assignment_arguments():
                self.inputs = self.build_assignment_inputs(lam)
                is_to_add_data = not lam.atomic
            else:
                self.inputs = self.build_conditional_inputs()
                is_to_add_data = False
            self.outputs = self.build_outputs(lam)
            is_to_add_data &= len(self.outputs) == 0

        if self.equality_scope is not None and len(self.equalities) > 0:
            return JoinEqualityEvaluationExpr(lam, self.equality_scope, self.equalities, self.inputs, self.outputs)\
                .compile(context)

        if is_to_add_data:
            return AddDataEvaluationExpr(lam, self.inputs, self.variable is not None).compile(context)
        elif isinstance(self.inputs, dict):
            if self.projection is not None and self.projection.num == 1:
                output_projection = self.projection.variables[0].name
            else:
                output_projection = None
            if lam.atomic:
                return AtomicEvaluationExpr(lam, self.inputs, output_projection).compile(context)
            elif len(self.inputs) == 0:
                if isinstance(self.variable, ColumnVariable) or isinstance(self.variable, UnquoteVariable):
                    self.variable.output_projection = output_projection
                    return self.variable
                elif self.projection is not None and len(self.outputs) == 1:
                    return RetrieveAllDataExpr(lam, output_projection).compile(context)
                else:
                    return RetrievePartialDataExpr(lam, self.outputs).compile(context)
        self.eval_lam = lam
        from norm.models import Variable
        if len(self.outputs) > 0:
            variables = [Variable(self.outputs[v.name], v.type_) for v in self.eval_lam.variables
                         if v.name in self.outputs.keys()]
        else:
            variables = self.eval_lam.variables
        self.lam = context.temp_lambda(variables)
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        for jv in self.join_variables:
            jv.execute(context)
        df = self.eval_lam.data.query(self.inputs, engine='python')

        if isinstance(df, DataFrame) and len(self.outputs) > 0:
            df = df[self.outputs.keys()].rename(columns=self.outputs)
        self.lam.data = df
        return df


class JoinEqualityEvaluationExpr(NormExpression):

    def __init__(self, lam, scope, equalities, condition, outputs):
        """
        Join the input scope first and then evaluate
        Event(name?, ip?eip) & Cluster(ip=eip, name~'App'?cluster_name);
        :param lam: the current scope
        :param scope: the input scope
        :param equalities: the equalities, e.g., {'eip':'ip'}
        :param condition: the other condition, e.g., 'name.str.contains("App")'
        :param outputs: the outputs for the current scope, e.g., {'name':'cluster_name'}
        """
        super().__init__()
        self.eval_lam = lam
        self.scope = scope
        self.equalities = equalities
        if isinstance(condition, str) and len(condition.strip()) > 0:
            self.condition = condition
        else:
            self.condition = None
        self.outputs = outputs

    def compile(self, context):
        scope = self.scope
        eval_lam_name = self.eval_lam.name
        from norm.models import Variable
        if self.outputs is not None and len(self.outputs) > 0:
            for v in self.eval_lam.variables:
                orig_name = self.outputs.get(v.name, None)
                if orig_name == v.name:
                    self.outputs[v.name] = self.VARIABLE_SEPARATOR.join([eval_lam_name, v.name])
                    from norm.engine import QuantifiedLambda
                    if isinstance(self.scope, QuantifiedLambda):
                        i = -1
                        for i, col in enumerate(self.scope.cols):
                            if col == orig_name:
                                break
                        if i >= 0:
                            self.scope.cols[i] = self.outputs[v.name]

        else:
            self.outputs = OrderedDict((v.name, self.VARIABLE_SEPARATOR.join([eval_lam_name, v.name]))
                                       for v in self.eval_lam.variables)

        output_keys = set(self.outputs.keys())
        variables = [Variable(v.name, v.type_) for v in scope.variables] + \
                    [Variable(self.outputs[v.name], v.type_) for v in self.eval_lam.variables
                     if v.name in output_keys]
        self.lam = context.temp_lambda(variables)
        self.lam.cloned_from = scope
        return self

    def prepare_to_merge(self):
        if self.condition:
            to_merge = self.eval_lam.data.query(self.condition, engine='python')
        else:
            to_merge = self.eval_lam.data
        cols = list(self.outputs.keys()) + [rc for lc, rc in self.equalities.items() if rc not in self.outputs]
        return to_merge[cols].rename(columns=self.outputs)

    def execute(self, context):
        from norm.engine import QuantifiedLambda
        if not isinstance(self.lam.cloned_from, QuantifiedLambda):
            inp = self.lam.cloned_from.data
        else:
            inp = self.lam.cloned_from.execute(context)
        if isinstance(inp, (DataFrame, Series)):
            if inp.index.name == self.lam.VAR_OID:
                inp = inp.reset_index()
        elif isinstance(inp, Index):
            inp = DataFrame(data=inp)
        equal_cols = list(self.equalities.items())
        left_col, right_col = equal_cols.pop()
        to_merge = self.prepare_to_merge()
        joined = inp.merge(to_merge, left_on=left_col, right_on=self.outputs.get(right_col, right_col), how='left')
        if right_col not in self.outputs:
            joined = joined.drop(columns=[right_col])
        if self.lam.VAR_OID not in joined.columns:
            joined.index.name = self.lam.VAR_OID
        else:
            joined = joined.set_index(self.lam.VAR_OID)
        condition = ' & '.join('({} == {})'.format(left_col, self.outputs.get(right_col, right_col))
                               for left_col, right_col in equal_cols)
        if condition != '':
            results = joined.query(condition)
        else:
            results = joined

        results = results.drop(columns=[right_col for left_col, right_col in equal_cols
                                        if right_col not in self.outputs])
        self.lam.data = results
        return results


class AtomicEvaluationExpr(NormExpression):

    def __init__(self, lam, inputs, output_projection=None):
        """
        Evaluate the atomic function
        :param lam: the function to evaluate
        :param inputs: the input values
        :param output_projection: the output projection
        """
        super().__init__()
        from norm.models.norm import Lambda
        assert(lam is not None)
        assert(isinstance(lam, Lambda))
        assert(isinstance(inputs, dict))
        self.eval_lam: Lambda = lam
        self.lam: Lambda = None
        self.inputs: Dict = inputs
        self.output_projection: str = output_projection

    def __str__(self):
        return '{}({})'.format(self.eval_lam.signature, ', '.join('{}={}'.format(key, value) for key, value
                                                                  in self.inputs.items()))

    def compile(self, context):
        from norm.models import Variable, Lambda
        self.lam = context.temp_lambda([Variable(Lambda.VAR_OUTPUT, self.eval_lam.output_type)])
        if self.output_projection is None and self.projection is not None:
            self.output_projection = self.projection.variables[0].name
        return self

    def execute(self, context):
        inputs = dict((key, value.execute(context)) for key, value in self.inputs.items())
        result = self.eval_lam(**inputs)
        import numpy
        if self.output_projection is not None:
            if isinstance(result, (list, numpy.ndarray)):
                return Series(result, name=self.output_projection)
            elif isinstance(result, (Series, Index)):
                return result.rename(self.output_projection)
            elif isinstance(result, DataFrame):
                if len(result.columns) > 1:
                    cols = {col: '{}{}{}'.format(self.output_projection, self.VARIABLE_SEPARATOR, col)
                            for col in result.columns}
                    return result.loc[result.index.rename(self.output_projection)].rename(columns=cols)
                else:
                    result.columns = [self.output_projection]
                    return result
            else:
                return Series([result], name=self.output_projection)
        else:
            if isinstance(result, DataFrame) and len(result.columns) == 1:
                from norm.models import Lambda
                result.columns = [Lambda.VAR_OUTPUT]
            return result


class RetrievePartialDataExpr(NormExpression):

    def __init__(self, lam, outputs):
        """
        Retrieve all data from the given Lambda
        :param lam: the given Lambda
        :param outputs: the output projections
        """
        super().__init__()
        from norm.models.norm import Lambda
        assert(lam is not None)
        assert(isinstance(lam, Lambda))
        assert(not lam.atomic)
        self.eval_lam: Lambda = lam
        self.lam: Lambda = None
        self.outputs: Dict = outputs if outputs is not None else OrderedDict()

    def __str__(self):
        oid = self.eval_lam.VAR_OID
        out = self.eval_lam.VAR_OUTPUT
        return '{}({}){}'.format(self.eval_lam.signature,
                                 ', '.join(('{}?'.format(c) for c in self.outputs.keys() if c != oid and c != out)),
                                 '?' if self.outputs.get(oid) or self.outputs.get(out) else '')

    def _append_projection(self):
        if self.projection is not None:
            assert(len(self.projection.variables) <= 1)
            if self.projection.num == 1:
                new_lambda_name = self.projection.variables[0].name
            else:
                new_lambda_name = self.eval_lam.name
            outputs = OrderedDict((key, self.VARIABLE_SEPARATOR.join([new_lambda_name, value])) for key, value
                                  in self.outputs.items())
            if self.eval_lam.is_functional:
                outputs[self.eval_lam.VAR_OUTPUT] = new_lambda_name
            else:
                outputs[self.eval_lam.VAR_OID] = new_lambda_name
            self.outputs = outputs
            self.projection = None

    def compile(self, context):
        from norm.models import Lambda, Variable
        self._append_projection()

        if len(self.outputs) > 0:
            if Lambda.VAR_OID in self.outputs.keys():
                from norm.models import lambdas
                variables = [Variable(self.outputs[Lambda.VAR_OID], lambdas.Integer)]
            else:
                variables = []
            variables += [Variable(self.outputs[v.name], v.type_) for v in self.eval_lam.variables
                          if v.name in self.outputs.keys()]
        else:
            variables = self.eval_lam.variables
        self.lam = context.temp_lambda(variables)
        return self

    def execute(self, context):
        oid_output_col = self.outputs.get(self.eval_lam.VAR_OID)
        cols = [col for col in self.outputs.keys() if col != self.eval_lam.VAR_OID]

        result = self.eval_lam.data[cols].rename(columns=self.outputs)

        if oid_output_col is not None:
            result.index.rename(oid_output_col, inplace=True)
            result = result.reset_index()
        self.lam.data = result
        return result


class RetrieveAllDataExpr(NormExpression):

    def __init__(self, lam, output_projection=None):
        """
        Retrieve all data from the given Lambda
        :param lam: the given Lambda
        :param output_projection: the output projection variable name
        """
        super().__init__()
        from norm.models.norm import Lambda
        assert(lam is not None)
        assert(isinstance(lam, Lambda))
        assert(not lam.atomic)
        self.eval_lam: Lambda = lam
        self.lam: Lambda = lam
        self.output_projection: str = output_projection

    def __str__(self):
        return '{}?'.format(self.lam.signature)

    def compile(self, context):
        if self.output_projection is not None:
            from norm.models.norm import Variable
            from norm.models import lambdas
            self.lam = context.temp_lambda([Variable(self.output_projection, lambdas.Integer)])
        return self

    def execute(self, context):
        if self.eval_lam.is_functional:
            result = self.eval_lam.data[self.lam.VAR_OUTPUT]
        else:
            result = self.eval_lam.data.index
        if self.output_projection is not None:
            return result.rename(self.output_projection)
        else:
            return result


class AddDataEvaluationExpr(NormExpression):

    def __init__(self, lam, data, immediately=True):
        """
        Add data to a Lambda. If the lambda is not the current scope, revision occurs immediately. Otherwise,
        revision is delayed to TypeImplementation
        :param lam: the Lambda to add data
        :type lam: norm.model.norm.Lambda
        :param data: the data to add
        :type data: Dict
        :param immediately: whether to revise Lambda as a disjunction immediately, default to True
        :type immediately: bool
        """
        super().__init__()
        from norm.models.norm import Lambda
        self.lam: Lambda = lam
        self.data: Dict = data
        self.immediately: bool = immediately
        self.description: str = 'add data'
        self._query_str: str = None

    def __str__(self):
        if self._query_str is None:
            msg = "Need to execute first!"
            logger.error(msg)
            raise NormError(msg)
        return self._query_str

    def compile(self, context):
        if not self.immediately and len(self.data) == 1:
            expr = list(self.data.values())[0]
            from norm.executable.expression.arithmetic import ArithmeticExpr
            if isinstance(expr, ArithmeticExpr) and expr.op is not None:
                return expr
        return self

    def execute(self, context):
        if len(self.data) == 0:
            df = DataFrame(data=[self.lam.default])
            self._query_str = hash_df(df)
            return df

        from norm.executable.constant import Constant
        from norm.executable import NormExecutable
        data = OrderedDict((key, value.execute(context) if isinstance(value, (Constant, NormExecutable)) else value)
                           for key, value in self.data.items())
        data = self.unify(data)
        df = DataFrame(data=data, columns=data.keys())
        query_str = hash_df(df)
        self._query_str = query_str
        df = self.lam.fill_primary(df)
        df = self.lam.fill_time(df)
        df = self.lam.fill_oid(df)
        if self.lam.VAR_OID in df.columns:
            df = df.set_index(self.lam.VAR_OID)
        else:
            df.index.rename(self.lam.VAR_OID, inplace=True)
        if self.immediately:
            from norm.models.norm import RevisionType
            df = self.lam.revise(query_str, self.description, df, RevisionType.DISJUNCTION)
            if self.lam.is_functional:
                return df[self.lam.VAR_OUTPUT]
            else:
                return df.index
        return df


class ChainedEvaluationExpr(NormExpression):

    def __init__(self, lexpr, rexpr):
        """
        Chained evaluation expressions:
        Example:
        * a.b
        * a.test(sf, s)
        * a.b.c
        :param lexpr: base query expressions or chained expression
        :type lexpr: Union[VariableName, EvaluationExpr]
        :param rexpr: chained evaluation expression
        :type rexpr: Union[EvaluationExpr, VariableName]
        """
        super().__init__()
        self.lexpr = lexpr
        self.rexpr = rexpr

    def compile(self, context):
        if isinstance(self.rexpr, VariableName):
            if isinstance(self.lexpr, VariableName):
                return VariableName(self.lexpr, self.rexpr.name).compile(context)
            elif isinstance(self.lexpr, EvaluationExpr):
                if self.rexpr.name in self.lexpr.lam:
                    return ColumnVariable(self.lexpr, self.rexpr.name)
                else:
                    msg = '{} does not exist in the {}'.format(self.rexpr, self.lexpr)
                    logger.error(msg)
                    raise NormError(msg)
            else:
                msg = 'Only a variable or an evaluation can have chained operation for now, ' \
                      'but got {}'.format(self.lexpr)
                logger.error(msg)
                raise NormError(msg)
        elif isinstance(self.rexpr, EvaluationExpr):
            if self.rexpr.lam is None and isinstance(self.lexpr, ColumnVariable):
                return DataFrameColumnFunctionExpr(self.lexpr, self.rexpr).compile(context)
            else:
                # Evaluation with the previous expression as the first input argument
                self.rexpr.args = [ArgumentExpr(expr=self.lexpr)] + self.rexpr.args
                # Recompile the expression
                return self.rexpr.compile(context)
        else:
            msg = 'Only chaining on a variable or an evaluation, but got {}'.format(self.rexpr)
            logger.error(msg)
            raise NormError(msg)


class DataFrameColumnFunctionExpr(EvaluationExpr):

    AGG_FUNCS = {'abs', 'idxmax', 'idxmin', 'count', 'kurt', 'mad', 'max', 'min', 'median', 'mean', 'nunique', 'std',
                 'sum'}
    INT_FUNCS = {'idxmax', 'idxmin', 'count', 'nunique'}
    FLT_FUNCS = {'mean', 'std', 'kurt'}
    STR_FUNCS = {'capitalize', 'cat', 'pcount', 'contains', 'endswith', 'extract', 'extractall', 'find', 'findall',
                 'len', 'lower', 'match', 'replace', 'split', 'strip', 'title', 'upper', 'isalnum', 'isalpha', 'isdigit',
                 'isspace', 'islower', 'isupper', 'istitle', 'isnumeric', 'isdecimal', 'wrap', 'translate', 'swapcase',
                 'slice', 'slice_replace', 'repeat', 'pad', 'startswith', 'normalize'}

    def __init__(self, column_variable, expr, projection=None):
        super().__init__([])
        self.column_variable: ColumnVariable = column_variable
        self.expr: EvaluationExpr = expr
        self.projection: Projection = projection

    def __str__(self):
        return '{}{}{}'.format(self.column_variable.name, self.VARIABLE_SEPARATOR, self.expr.variable.name)

    @property
    def var_name(self):
        if self.projection is not None:
            assert(len(self.projection.variables) == 1)
            return self.projection.variables[0].name
        else:
            return str(self)

    def compile(self, context):
        self.eval_lam = self.column_variable.lam
        from norm.models.norm import Variable
        from norm.engine import QuantifiedLambda
        if isinstance(self.column_variable.lam, QuantifiedLambda):
            variables = self.column_variable.lam.get_grouped_variables()
        else:
            variables = []
        variables.append(Variable(self.var_name, self.var_type))
        self.lam = context.temp_lambda(variables=variables)
        self.lam.cloned_from = self.eval_lam
        return self

    @property
    def is_aggregation(self):
        return self.expr.variable.name in self.AGG_FUNCS

    @property
    def is_str_handling(self):
        return self.expr.variable.name in self.STR_FUNCS

    @property
    def var_type(self):
        from norm.models import lambdas
        f = self.expr.variable.name
        if self.is_aggregation:
            if f in self.INT_FUNCS:
                return lambdas.Integer
            if f in self.FLT_FUNCS:
                return lambdas.Float
            return self.column_variable.variable_type()
        if self.is_str_handling:
            return lambdas.String

    def get_args(self, context):
        args = []
        kwargs = {}
        for arg in self.expr.args:  # type: ArgumentExpr
            if arg.expr is None:
                continue
            if arg.op is None and arg.variable is not None:
                kwargs[arg.variable.name] = arg.expr.execute(context)
            else:
                args.append(arg.expr.execute(context))
        return args, kwargs

    def execute(self, context):
        args, kwargs = self.get_args(context)
        col = self.column_variable.execute(context)
        f = self.expr.variable.name
        if self.is_str_handling:
            df = getattr(col.str, f)(*args, **kwargs)
            if not isinstance(df, Series):
                df = Series(data=[df], name=self.var_name, dtype=self.var_type.dtype)
            else:
                df.name = self.var_name
        elif self.is_aggregation:
            df = getattr(col, f)(*args, **kwargs)
            from norm.engine import QuantifiedLambda
            if isinstance(self.column_variable.lam, QuantifiedLambda):
                if isinstance(df, Series):
                    df = DataFrame(data={self.var_name: df}).reset_index()
                else:
                    df = DataFrame(data={self.var_name: df}, index=col.index).reset_index()
            else:
                df = Series(data=[df], name=self.var_name, dtype=self.var_type.dtype)
        else:
            msg = '{} is not implemented yet'.format(self.expr.variable.name)
            logging.error(msg)
            raise NormError(msg)
        self.lam.data = df
        return df



from collections import OrderedDict
from typing import List

from pandas import DataFrame, Series, merge, concat, Index

from norm.executable import NormError
from norm.executable.constant import ListConstant, Constant, TupleConstant
from norm.executable.expression import NormExpression
from norm.executable.expression.condition import ConditionExpr, CombinedConditionExpr
from norm.executable.expression.evaluation import AddDataEvaluationExpr, DataFrameColumnFunctionExpr
from norm.grammar.literals import LOP

import logging
logger = logging.getLogger(__name__)

CROSS_JOIN_KEY = '__key__'


class QueryExpr(NormExpression):

    def __init__(self, op, expr1, expr2):
        """
        Query expression
        :param op: logical operation, e.g., [&, |, ^, =>, <=>]
        :type op: LOP
        :param expr1: left expression
        :type expr1: NormExpression
        :param expr2: right expression
        :type expr2: NormExpression
        """
        super().__init__()
        self.op = op
        self.expr1 = expr1
        self.expr2 = expr2
        self.cross_join = False
        self.common_columns = []

    def __combine_value(self, value1, value2):
        if value1 is None or value2 is None:
            return None
        elif isinstance(value1, ListConstant) and isinstance(value2, ListConstant):
            value1.value.extend(value2.value)
            return value1
        elif isinstance(value1, ListConstant) and isinstance(value2, Constant):
            value1.value.append(value2.value)
            return value1
        elif isinstance(value1, Constant) and isinstance(value2, ListConstant):
            value2.value.append(value1.value)
            return value2
        elif isinstance(value1, Constant) and isinstance(value2, Constant):
            return ListConstant(value1.type_, [value1.value, value2.value])
        elif isinstance(value1, list) and isinstance(value2, list):
            value1.extend(value2)
            return value1
        elif isinstance(value1, list) and not isinstance(value2, list):
            value1.append(value2)
            return value1
        elif not isinstance(value1, list) and isinstance(value2, list):
            value2.append(value1)
            return value2
        elif isinstance(value1, AddDataEvaluationExpr) and isinstance(value2, AddDataEvaluationExpr):
            # pushdown the combine operation
            if value1.lam is value2.lam:
                combined = self.__combine_data(value1.data, value2.data)
                if combined is not None:
                    value1.data = combined
                    return value1
        elif not isinstance(value1, list) and not isinstance(value2, list):
            return [value1, value2]
        return None

    def __combine_data(self, data1, data2):
        cols = list(data1.keys())
        cols.extend(col for col in data2.keys() if col not in data1.keys())
        data = OrderedDict()
        for col in cols:
            combined = self.__combine_value(data1.get(col), data2.get(col))
            if combined is None:
                return None
            else:
                data[col] = combined
        return data

    def __same_origin(self):
        """
        Decide whether two expressions come from the same origin Lambda. Check the transformation path between two
        expressions. If there is a node that overlaps, the two expressions have the same origin.
        :return: True if they are
        :rtype: bool
        """
        if self.expr1.eval_lam is self.expr2.eval_lam:
            return True
        if self.expr1.eval_lam.cloned_from is self.expr2.eval_lam.cloned_from:
            return True
        if self.expr1.lam is self.expr2.eval_lam:
            return True
        if self.expr1.lam is self.expr2.scope:
            return True

    def compile(self, context):
        if isinstance(self.expr1, AddDataEvaluationExpr) and isinstance(self.expr2, AddDataEvaluationExpr)\
                and self.expr1.lam is self.expr2.lam:
            # Combine the constant arguments for adding data
            # ensure that all arguments are constants
            combined = self.__combine_data(self.expr1.data, self.expr2.data)
            if combined is not None:
                self.expr1.data = combined
                return self.expr1
        if isinstance(self.expr1, ConditionExpr) and isinstance(self.expr2, ConditionExpr) and \
                (self.expr1.eval_lam is self.expr2.eval_lam or self.expr1.eval_lam is self.expr2.eval_lam.cloned_from):
            return CombinedConditionExpr(self.op, self.expr1, self.expr2).compile(context)

        self.lam = context.temp_lambda(self.expr1.lam.variables +
                                       [v for v in self.expr2.lam.variables if v.name not in self.expr1.lam])
        self.eval_lam = self.expr1.eval_lam
        from norm.executable.schema.variable import VariableName
        if isinstance(self.expr2, VariableName) and not isinstance(self.expr1, VariableName):
            self.common_columns = [self.expr2.name] if self.expr2.name in self.expr1.lam else []
        elif not isinstance(self.expr2, VariableName) and isinstance(self.expr1, VariableName):
            self.common_columns = [self.expr1.name] if self.expr1.name in self.expr2.lam else []
        elif isinstance(self.expr2, VariableName) and isinstance(self.expr1, VariableName):
            self.common_columns = [self.expr1.name] if self.expr1.name == self.expr2.name else []
        else:
            self.common_columns = [v.name for v in self.expr1.lam.variables if v.name in self.expr2.lam]
        self.cross_join = not self.__same_origin() and len(self.common_columns) == 0
        return self

    def execute(self, context):
        df1 = self.expr1.execute(context)
        if isinstance(df1, Series):
            df1 = DataFrame(data={df1.name: df1})
        df2 = self.expr2.execute(context)
        if isinstance(df2, Series):
            df2 = DataFrame(data={df2.name: df2})

        # TODO: OR to union
        from norm.models.norm import Lambda
        if self.op == LOP.AND:
            # both indices are the same, left join to df2
            # if not, cross join df1 and df2
            if self.cross_join:
                # TODO: check the logic
                df1[CROSS_JOIN_KEY] = 1
                df2[CROSS_JOIN_KEY] = 1
                if df1.index.name is not None:
                    df1 = df1.reset_index()
                if df2.index.name is not None:
                    df2 = df2.reset_index()
                cols = [col for col in df1.columns if col not in df2.columns]
                df2 = merge([df2, df1[cols]]).drop(columns=[CROSS_JOIN_KEY])
            else:
                if len(self.common_columns) == 0:
                    cols = [col for col in df1.columns if col not in df2.columns]
                    if len(cols) > 0:
                        df2[cols] = df1.loc[df2.index, cols]
                else:
                    # preserve df1 index if available
                    if isinstance(df1, Index):
                        df1 = self.expr1.lam.data.loc[df1]
                    if isinstance(df2, Index):
                        df2 = self.expr2.lam.data.loc[df2, [v.name for v in self.expr2.lam.variables]]
                    if df1.index.name == self.lam.VAR_OID:
                        df1 = df1.reset_index()
                    for col in self.common_columns:
                        if df1[col].dtype != df2[col].dtype:
                            df2[col] = df2[col].astype(df1[col].dtype)
                    df2 = merge(df1, df2, on=self.common_columns).dropna()
                    if self.lam.VAR_OID in df2.columns:
                        df2 = df2.set_index(self.lam.VAR_OID)

        elif self.op == LOP.OR:
            if len(self.common_columns) > 0:
                df2 = merge(df1, df2, on=self.common_columns, how='outer')
            else:
                df2 = concat([df1, df2])

        if df2.index.name != Lambda.VAR_OID and df2.index.name is not None:
            df2 = df2.reset_index()
        self.lam.data = df2
        return df2


class ConstantAssignmentExpr(NormExpression):

    def __init__(self, constant, projection=None):
        """
        Assign constant to a variable and join with the current scope
        :param constant: the constant to assign
        :type constant: Constant
        :param projection: the variable to project to
        :type projection: Projection
        """
        super().__init__()
        from norm.executable import Projection
        self.constant: Constant = constant
        self.projection: Projection = projection
        self.variable_name: str = None

    @property
    def var_type(self):
        from norm.models import lambdas
        return lambdas[self.constant.type_.value.title()]

    def compile(self, context):
        assert(self.projection is not None)
        if len(self.projection.variables) > 1:
            return MultipleConstantAssignmentExpr(self.constant, self.projection).compile(context)

        from norm.models.norm import Variable
        self.variable_name = self.projection.variables[0].name
        if context.scope is not None:
            self.eval_lam = context.scope
            variables = self.eval_lam.variables
            self.lam = context.temp_lambda(variables + [Variable(self.variable_name, self.var_type)])
        else:
            self.eval_lam = None
            self.lam = context.temp_lambda([Variable(self.variable_name, self.var_type)])
        return self

    def execute(self, context):
        if not isinstance(self.constant, ListConstant):
            if self.eval_lam is not None:
                df = self.eval_lam.data
                df[self.variable_name] = self.constant.value
            else:
                df = DataFrame(data={self.variable_name: [self.constant.value]})
        else:
            if self.eval_lam is not None:
                df = self.eval_lam.data
                if self.variable_name not in df.columns:
                    df[CROSS_JOIN_KEY] = 1
                    to_join = DataFrame({self.variable_name: self.constant.value})
                    to_join[CROSS_JOIN_KEY] = 1
                    df = df.merge(to_join, on=CROSS_JOIN_KEY, how='outer').drop(columns=[CROSS_JOIN_KEY])
                else:
                    assert(len(df) == len(self.constant.value))
                    df[self.variable_name] = self.constant.value
            else:
                df = DataFrame(data={self.variable_name: self.constant.value})
        self.lam.data = df
        return df


class MultipleConstantAssignmentExpr(NormExpression):

    def __init__(self, constant, projection):
        """
        Assign constant to a variable and join with the current scope
        :param constant: the constant to assign
        :type constant: Constant
        :param projection: the variables to project to
        :type projection: Projection
        """
        super().__init__()
        from norm.executable import Projection
        self.constant: Constant = constant
        self.projection: Projection = projection

    @property
    def var_names(self):
        return [v.name for v in self.projection.variables]

    @property
    def var_types(self):
        from norm.models import lambdas
        return [lambdas[t.value.title()] for t in self.constant.type_]

    def compile(self, context):
        assert(isinstance(self.constant, ListConstant))
        assert(isinstance(self.constant.value[0], tuple))
        assert(len(self.var_names) == len(self.var_types))
        assert(self.eval_lam is None or all(vn not in self.eval_lam for vn in self.var_names))
        from norm.models.norm import Variable
        vs = [Variable(n, t) for n, t in zip(self.var_names, self.var_types)]
        if context.scope is not None:
            self.eval_lam = context.scope
            variables = self.eval_lam.variables
            self.lam = context.temp_lambda(variables + vs)
        else:
            self.eval_lam = None
            self.lam = context.temp_lambda(vs)
        return self

    def execute(self, context):
        if self.eval_lam is not None:
            df = self.eval_lam.data
            if any(vn in df.columns for vn in self.var_names):
                assert(len(df) == len(self.constant.value))
                df[self.var_names] = self.constant.value
            else:
                df[CROSS_JOIN_KEY] = 1
                to_join = DataFrame(data=self.constant.value,
                                    columns=self.var_names)
                to_join[CROSS_JOIN_KEY] = 1
                df = df.merge(to_join, on=CROSS_JOIN_KEY, how='outer').drop(columns=[CROSS_JOIN_KEY])
        else:
            df = DataFrame(data=self.constant.value, columns=self.var_names)
        self.lam.data = df
        return df


class QuantifiedQueryExpr(NormExpression):

    def __init__(self, expr, quantifiers):
        """
        Quantify result by the quantifiers
        :param expr: the expression to query
        :type expr: NormExpression
        :param quantifiers: the list of quantifiers to evaluate
        :type quantifiers: Quantifiers
        """
        super().__init__()
        self.expr = expr
        self.quantifiers = quantifiers

    def compile(self, context):
        self.eval_lam = self.expr.eval_lam
        self.lam = self.expr.lam
        return self

    def execute(self, context):
        self.quantifiers.execute(context)
        df = self.expr.execute(context)
        assert(isinstance(df, DataFrame))
        df = self.quantifiers.quantify(df)
        from norm.models.norm import Lambda
        if isinstance(self.lam, Lambda):
            self.lam.data = df
        return df


class PivotQueryExpr(NormExpression):

    def __init__(self, expr, projection):
        """
        Unquoted projection essentially does the pivoting
        :param expr: the expression to evaluate
        :type expr: NormExpression
        :param projection: the unquoted projection to pivot
        :type projection: Projection
        """
        super().__init__()
        self.expr = expr
        self.projection = projection
        self.cols = []
        self.var_name = None

    def compile(self, context):
        assert(len(self.projection.variables) == 1)
        from norm.executable.schema.variable import UnquoteVariable
        v = self.projection.variables[0]
        assert(isinstance(v, UnquoteVariable))
        self.cols = [uv.name for uv in v.unquoted_variables]
        self.var_name = v.name
        self.lam = self.expr.lam
        self.eval_lam = v.lam
        return self

    def execute(self, context):
        result = self.expr.execute(context)
        # TODO: this is a hack, have to think about which value columns in general are referred to
        col_values = [self.expr.lam.variables[-1].name]
        col_index = [col for col in result.columns if col not in col_values and col not in self.cols]
        if len(col_index) == 0:
            col_index = result.index
        for col in self.cols:
            if col not in result.columns:
                result[col] = self.eval_lam.data[col]
        df = result.pivot_table(values=col_values, columns=self.cols, index=col_index, aggfunc=lambda x: x)
        df.columns = [self.var_name.format(*col_values) for col_values in zip(*df.columns.levels[1:])]
        if col_index is not result.index:
            df = df.reset_index()
        return df


class NegatedQueryExpr(NormExpression):

    def __init__(self, expr):
        """
        Negation of the expression
        :param expr: the expression to negate
        :type expr: NormExpression
        """
        super().__init__()
        self.expr = expr

    def compile(self, context):
        if isinstance(self.expr, QueryExpr):
            self.expr.expr1 = NegatedQueryExpr(self.expr.expr1).compile(context)
            self.expr.expr2 = NegatedQueryExpr(self.expr.expr2).compile(context)
            self.expr.op = self.expr.op.negate()
        elif isinstance(self.expr, ConditionExpr):
            self.expr.op = self.expr.op.negate()
        else:
            msg = 'Currently NOT only works on logically combined query or conditional query'
            logger.error(msg)
            raise NotImplementedError(msg)
        return self.expr

    def execute(self, context):
        msg = 'Negated query is not executable, but only compilable'
        logger.error(msg)
        raise NormError(msg)




from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.pool import StaticPool

from IPython import get_ipython
from IPython.core.magic import register_line_magic, register_cell_magic, register_line_cell_magic

import os
import logging
logger = logging.getLogger(__name__)

context = None


def configure(home=None, db_path=None, data_path=None, **kwargs):
    """
    Setting the parameters for Norm
    :param home: the path to the db and the data, e.g., /home/<user>/.norm
    :type home: str
    :param db_path: the path to the db, e.g., /home/<user>/.norm/db/norm.db
    :type db_path: str
    :param data_path: the path to the data, e.g., /home/<user>/.norm/data
    :type data_path: str
    :param kwargs: other parameters
    """
    from norm.config import Session
    from norm import config

    if home is not None:
        config.NORM_HOME = home
        config.DATA_STORAGE_ROOT = os.path.join(home, 'data')
        config.DB_PATH = os.path.join(home, 'db/norm.db')
    if db_path is not None:
        config.DB_PATH = db_path
    if data_path is not None:
        config.DATA_STORAGE_ROOT = data_path
    if home is not None or db_path is not None:
        config.engine = create_engine('sqlite:///{}'.format(config.DB_PATH), poolclass=StaticPool)
        Session.configure(bind=config.engine)
        config.session = Session()
        config.context_id = str(datetime.utcnow().strftime('%m%d%Y.%H%M%S'))
        global context
        from norm.security import user
        from norm.engine import NormCompiler, NormError
        context = NormCompiler(config.context_id, user, config.session)


def init_jupyter():
    from norm.engine import NormCompiler, NormError
    from norm.config import session, context_id
    from norm.security import user, login
    logging.disable()
    global context
    user = login()
    context = NormCompiler(context_id, user, session)


def init_colab():
    """
    Setting the configurations in Colab (Google)
    """
    # Login the user
    from google.colab import auth
    from google.colab import drive
    auth.authenticate_user()
    import requests
    from gcloud import credentials
    logging.basicConfig(level=logging.DEBUG)
    access_token = credentials.get_credentials().get_access_token().access_token
    gcloud_tokeninfo = requests.get('https://www.googleapis.com/oauth2/v3/userinfo?access_token=' + access_token).json()
    email = gcloud_tokeninfo['email']
    last_name = gcloud_tokeninfo['hd']
    first_name = email.split('@')[0]

    # Mount the drive
    drive.mount('/content/drive')

    # Configure the database
    home = '/content/drive/My Drive/.norm/'
    if not os.path.exists(home):
        os.mkdir(home)
        logging.info("Directory " + home + " created ")
    else:
        logging.info("Directory " + home + " already exists")

    from norm.config import Session
    from norm import config, security

    config.NORM_HOME = home

    config.DATA_STORAGE_ROOT = os.path.join(home, 'data')
    if not os.path.exists(config.DATA_STORAGE_ROOT):
        os.mkdir(config.DATA_STORAGE_ROOT)
        logging.info("Directory " + config.DATA_STORAGE_ROOT + " created ")
    else:
        logging.info("Directory " + config.DATA_STORAGE_ROOT + " already exists")

    db_path = os.path.join(home, 'db')
    if not os.path.exists(db_path):
        os.mkdir(db_path)
        logging.info("Directory " + db_path + " created ")
    else:
        logging.info("Directory " + db_path + " already exists")

    config.DB_PATH = os.path.join(home, 'db/norm.db')
    if not os.path.exists(config.DB_PATH):
        orig_db_file = os.path.join(os.path.dirname(__file__), 'db/norm.db')
        from shutil import copyfile
        copyfile(orig_db_file, config.DB_PATH)
        logging.info("File " + config.DB_PATH + " copied")
    else:
        logging.info("File " + config.DB_PATH + " already exists")

    config.engine = create_engine('sqlite:///{}'.format(config.DB_PATH), poolclass=StaticPool)
    Session.configure(bind=config.engine)
    config.session = Session()
    config.context_id = str(datetime.utcnow().strftime('%m%d%Y.%H%M%S'))

    from norm.security import login
    user = login({'first_name': first_name,
                  'last_name': last_name,
                  'username': email,
                  'email': email})

    logging.info(user.username + ' logged in')

    global context
    from norm.engine import NormCompiler, NormError
    context = NormCompiler(config.context_id, user, config.session)


# IPython magics
if get_ipython() is not None:
    def update_python_context():
        context.python_context = ip.user_global_ns

    def try_norma(lines):
        for i in range(len(lines) - 1, -1, -1):
            if lines[i].rstrip(' \t\r\n') != '':
                lines = lines[:(i+1)]
                break
        if lines[-1].rstrip(' \t\r\n')[-1] == ';':
            context.python_context = ip.user_global_ns
            return ['%%norma\n'] + lines
        else:
            return lines
    ip = get_ipython()
    try:
        ip.input_transformers_cleanup.append(try_norma)
    except:
        logging.warning("Input transformers are not available. Magic %norma or %%norma is required")

    @register_line_cell_magic
    def norma(line, cell=None):
        """
        Parsing the norm command and execute it
        :param line: a line of norm command
        :type line: str
        :param cell: a multi-line of norm command
        :type cell: str
        """
        from norm.engine import NormCompiler, NormError
        import norm.config as config
        try:
            if cell is None:
                result = context.execute(line)
            else:
                result = context.execute(cell)

            config.session.commit()
            return result
        except SQLAlchemyError as e:
            config.session.rollback()
            msg = 'Session commit failed on {}'.format(config.engine)
            logger.error(msg)
            logger.error(e)
            raise NormError(msg)



from typing import Union

from pandas.core.groupby import DataFrameGroupBy

from norm.executable import NormError
from norm.executable.constant import Constant
from norm.grammar.literals import COP, ConstantType, LOP

from norm.executable.expression import NormExpression
from norm.executable.schema.variable import VariableName

import logging
logger = logging.getLogger(__name__)


class ConditionExpr(NormExpression):

    def __init__(self, op, lexpr, rexpr):
        """
        Condition expression
        :param op: conditional operation, e.g., [<, <=, >, >=, =, !=, in, !in, ~]. ~ means 'like'
        :type op: Union[COP, LOP]
        :param lexpr: arithmetic expression, e.g., a + b - c
        :type lexpr: Union[ArithmeticExpr, ConditionExpr]
        :param rexpr: another arithmetic expression
        :type rexpr: Union[ArithmeticExpr, ConditionExpr]
        """
        super().__init__()
        from norm.executable.expression.arithmetic import ArithmeticExpr
        self.op: COP = op
        self.lexpr: Union[ArithmeticExpr, ConditionExpr] = lexpr
        self.rexpr: Union[ArithmeticExpr, ConditionExpr] = rexpr
        self._condition = None
        assert(self.lexpr is not None)
        assert(self.rexpr is not None)
        assert(self.op is not None)

    def __str__(self):
        if self._condition is None:
            msg = 'Compile the condition expression first'
            logger.error(msg)
            raise NormError(msg)
        return self._condition

    def compile(self, context):
        if self.op == COP.LK:
            assert(isinstance(self.lexpr, VariableName))
            assert(isinstance(self.rexpr, Constant))
            assert(self.rexpr.type_ is ConstantType.STR or self.rexpr.type_ is ConstantType.PTN)
            self._condition = '{}.str.contains({})'.format(self.lexpr, self.rexpr)
        else:
            self._condition = '({}) {} ({})'.format(self.lexpr, self.op, self.rexpr)
        self.eval_lam = self.lexpr.lam
        self.lam = context.temp_lambda(self.eval_lam.variables)
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        self.lexpr.execute(context)
        self.rexpr.execute(context)
        data = self.eval_lam.data
        if not isinstance(self.rexpr, Constant):
            additional_cols = [v.name for v in self.rexpr.lam.variables if v.name not in data.columns]
            if len(additional_cols) > 0:
                data[additional_cols] = self.rexpr.lam.data.set_index(self.lam.VAR_OID)
        if isinstance(data, DataFrameGroupBy):
            self.lam.data = data.apply(lambda x: x.query(self._condition, engine='python')).reset_index(drop=True)
        else:
            self.lam.data = data.query(self._condition, engine='python')
        return self.lam.data


class CombinedConditionExpr(ConditionExpr):

    def __init__(self, op, lexpr, rexpr):
        """
        Combined conditional expression
        :param op: logical operation, e.g., ['and', 'or'] others are not supported yet # TODO support xor, imp, eqv
        :type op: LOP
        :param lexpr: left conditional expression
        :type lexpr: ConditionExpr
        :param rexpr:right conditional expression
        :type rexpr: ConditionExpr
        """
        super().__init__(op, lexpr, rexpr)

    def compile(self, context):
        if self._condition is not None:
            return self

        self._condition = '({}) {} ({})'.format(self.lexpr, self.op, self.rexpr)
        assert(self.lexpr.eval_lam is self.rexpr.eval_lam or self.lexpr.eval_lam is self.rexpr.eval_lam.cloned_from)
        self.eval_lam = self.lexpr.eval_lam
        self.lam = context.temp_lambda(self.eval_lam.variables)
        self.lam.cloned_from = self.eval_lam
        return self

    def execute(self, context):
        data = self.eval_lam.data
        if isinstance(data, DataFrameGroupBy):
            self.lam.data = data.apply(lambda x: x.query(self._condition, engine='python'))\
                .reset_index(drop=True)
        else:
            self.lam.data = data.query(self._condition, engine='python')
        return self.lam.data
